#define QUANT
#include <stdlib.h>
#include <limits.h>
#include <Xm/Xm.h>
#include <string.h>
#include "paint.h"

/* function returns the distance of two colors */
DQUANT unsigned long getDistance(unsigned char *col1, unsigned char *col2)
{
 int dr, dg, db;

 dr = abs (col1[RED] - col2[RED]);
 dg = abs (col1[GREEN] - col2[GREEN]);
 db = abs (col1[BLUE] - col2[BLUE]);
 return(dr * dr + dg * dg + db * db);
}

/* init histogram for quantization */
DQUANT void initHist(unsigned long *colHist)
{
 memset(colHist, 0, sizeof (unsigned long) * MAX_CELLS);
}

/* compute histogram for quantization */
DQUANT void addToHist(unsigned long *colHist, unsigned char *data, int x, int y, int width, int height, int bpl)
{
 int i;
 unsigned char *pos, *limit;

 for (i = 0; i < height; i++)
  {
   pos = &data[(i + y) * bpl + x];
   limit = pos + width;
   while (pos != limit)
     colHist[*pos++]++;
  }
}

/* compute histogram for quantization from a region; dx and dy are offset of the 
region to the x and y*/
DQUANT void addToHistFromRegion(unsigned long *colHist, unsigned char *data,
 int x, int y, unsigned int width, unsigned int height, unsigned int bpl,
 Region region, int dx, int dy)
{
 int i, j, limitX, limitY;
 unsigned char *pos;

 if (region == NULL)
  {
   addToHist(colHist, data, x, y, width, height, bpl);
   return;
  }

 limitX = x + width;
 limitY = y + height;
 for (i = y; i < limitY; i++)
  {
   pos = &data[i * bpl + x];
   for (j = x; j < limitX; j++)
    {
     if (XPointInRegion(region, j, i))
      colHist[*pos]++;
     pos++;
    }
  }
}

/* quantize srcDstImage to requestNumCol colors, shuffle the XImage according */
/* to allocated colors in allocCells (their number must be >= requestNumCol) */
/* source colormap must be in srcCmap, resulting number of colors is returned */
/* in resultNumCol (may be smaller then requestNumCol but not bigger) */
DQUANT void quantize (XImage * srcDstImage, int requestNumCol, int *resultNumCol,
 unsigned char *srcCmap, XColor * allocCells, unsigned char *redCo, unsigned char *greenCo,
 unsigned char *blueCo, unsigned char *transTable, unsigned long *colHist)
{
 int i, j;
 int mindist, minval, minpos, dist, numColors, maxpos, maxval;
 Boolean eliminated[MAX_CELLS];
 unsigned long origHist[MAX_CELLS];

 for (i = 0; i < MAX_CELLS; i++)
   eliminated[i] = FALSE;

 /* preserve original histogram */
 for (i = 0; i < MAX_CELLS; i++)
   origHist[i] = colHist[i];

 /* eliminate duplicate colors */
 for (i = 0; i < MAX_CELLS; i++)
   if (colHist[i] > 0)
     for (j = i + 1; j < MAX_CELLS; j++)
       if (colHist[j] > 0)
	 if (getDistance (&srcCmap[i * 3], &srcCmap[j * 3]) == 0)
	   eliminated[j] = TRUE;

 /* count the actual number of colors */
 numColors = 0;
 for (i = 0; i < MAX_CELLS; i++)
   if ((colHist[i] > 0) && (!(eliminated[i])))
     numColors++;

 /* special case quantization for grayscale images to provide equidistant sampling */
 if (numColors > requestNumCol)
   {
     int grayscale = TRUE;
     int grayLevel[256], usedGray[256], remapGray[256];

     for (i = 0; i < MAX_CELLS; i++)
       if ((colHist[i] > 0) &&
	   ((srcCmap[i * 3] != srcCmap[i * 3 + 1]) || (srcCmap[i * 3 + 1] != srcCmap[i * 3 + 2])))
	 grayscale = FALSE;	/* non-empty non-gray color, ergo not a grayscale image */

     if (grayscale)
       {
	 for (i = 0; i < requestNumCol; i++)	/* setup gray levels */
	   grayLevel[i] = i * 255 / (requestNumCol - 1);

	 /* check which gray levels are really used */
	 for (i = 0; i < MAX_CELLS; i++)
	   usedGray[i] = FALSE;
	 for (i = 0; i < MAX_CELLS; i++)
	   if (colHist[i] > 0)
	     {
	       j = minpos = 0;
	       minval = 256;
	       for (j = 0; j < requestNumCol; j++)
		 if (abs (grayLevel[j] - (int) srcCmap[3 * i]) < minval)
		   {
		     minpos = j;
		     minval = abs (grayLevel[j] - (int) srcCmap[3 * i]);
		   }
	       usedGray[minpos] = TRUE;
	     }
	 /* set the remapping of gray levels to compress cmap, and cmap values */
	 numColors = 0;
	 for (i = 0; i < requestNumCol; i++)
	   if (usedGray[i])
	     {
	       redCo[numColors] = greenCo[numColors] = blueCo[numColors] = grayLevel[i];
	       allocCells[numColors].red = redCo[numColors] << 8;
	       allocCells[numColors].green = greenCo[numColors] << 8;
	       allocCells[numColors].blue = blueCo[numColors] << 8;
	       remapGray[i] = numColors++;
	     }
	 /* finally set the transformation array for remapping */
	 /* use the same function for finding closest gray as above */
	 for (i = 0; i < MAX_CELLS; i++)
	   if (colHist[i] > 0)
	     {
	       j = minpos = 0;
	       minval = 256;
	       for (j = 0; j < requestNumCol; j++)
		 if (abs (grayLevel[j] - (int) srcCmap[3 * i]) < minval)
		   {
		     minpos = j;
		     minval = abs (grayLevel[j] - (int) srcCmap[3 * i]);
		   }
	       transTable[(unsigned char) i] = (unsigned char)
		 allocCells[remapGray[minpos]].pixel;
	     }
	 /* do the actual remapping */
	 remapData ((unsigned char *) srcDstImage->data, transTable,
		    srcDstImage->bytes_per_line * srcDstImage->height);
	 *resultNumCol = numColors;
	 return;
       }
   }

 while (numColors > requestNumCol)
   {
     /* find non-eliminated color with minimum number of pixels */
     minval = INT_MAX;
     minpos = 0;
     for (i = 0; i < MAX_CELLS; i++)
       if ((colHist[i] > 0) && (!(eliminated[i])))
	 if (colHist[i] < minval)
	   {
	     minval = colHist[i];
	     minpos = i;
	   }
     eliminated[minpos] = TRUE;

     /* add some pixels to the closest colors histogram values */
     /* to improve its survival probability */
     mindist = INT_MAX;
     minval = 0;
     for (j = 0; j < MAX_CELLS; j++)
       if ((colHist[j] > 0) && (!(eliminated[j])))
	 {
	   dist = getDistance (&srcCmap[minpos * 3], &srcCmap[j * 3]);
	   if (dist < mindist)
	     {
	       mindist = dist;
	       minval = j;
	     }
	 }
     if (colHist[minval] < (INT_MAX >> 2))	/* this improves near grayscale conversion */
       colHist[minval] += (colHist[minpos] << 1);

     numColors--;
   }

 /* set colors and transTable for the remaining colors */
 /* in order of the number of pixels that they occupy */
 j = 0;
 for (i = 0; i < numColors; i++)
   {
     maxval = 0;
     maxpos = 0;
     for (j = 0; j < MAX_CELLS; j++)
       if ((origHist[j] > maxval) && (!(eliminated[j])))
	 {
	   maxval = origHist[j];
	   maxpos = j;
	 }
     transTable[(unsigned char) maxpos] = (unsigned char) allocCells[i].pixel;
     redCo[i] = srcCmap[3 * maxpos];
     greenCo[i] = srcCmap[3 * maxpos + 1];
     blueCo[i] = srcCmap[3 * maxpos + 2];
     allocCells[i].red = redCo[i] << 8;
     allocCells[i].green = greenCo[i] << 8;
     allocCells[i].blue = blueCo[i] << 8;

     origHist[maxpos] = 0;
   }

 /* find closest colors for the eliminated colors */
 for (i = 0; i < MAX_CELLS; i++)
   if ((colHist[i] > 0) && (eliminated[i]))
     {
       mindist = INT_MAX;
       minval = 0;
       for (j = 0; j < MAX_CELLS; j++)
	 if ((colHist[j] > 0) && (!(eliminated[j])))
	   {
	     dist = getDistance (&srcCmap[i * 3], &srcCmap[j * 3]);
	     if (dist < mindist)
	       {
		 mindist = dist;
		 minval = j;
	       }
	   }
       transTable[(unsigned char) i] = transTable[(unsigned char) minval];
     }
 remapData ((unsigned char *) srcDstImage->data, transTable,
	    srcDstImage->bytes_per_line * srcDstImage->height);
 *resultNumCol = numColors;
}

/* remaps the data passed according to values in transTable */
DQUANT void remapData(unsigned char *data, unsigned char *transTable, int length)
{
 unsigned char *limit;

 limit = &data[length];
 data--;
 while (++data < limit)
   *data = transTable[*data];
}

DQUANT void remapImgArea(XImage *img, unsigned char *transTable, int x,
 int y, unsigned int width, unsigned int height, Region region)
{
 unsigned char *ptr;
 int i, j;
 Rectangle imgR, reqR, resR;
 
 imgR.x = imgR.y = 0;
 imgR.width = img->width;
 imgR.height = img->height;
 reqR.x = x;
 reqR.y = y;
 reqR.width = width;
 reqR.height = height;

 if (intersectRect (&imgR, &reqR, &resR))
  {
   if (region == NULL) 
    for (i = resR.y; i < (resR.y + resR.height); i++)
     {
      ptr = &img->data[i*img->bytes_per_line + resR.x];
      remapData (ptr, transTable, resR.width);
     }
   else /* non rectangular area - check point by point */
    for (i = resR.y; i < (resR.y + resR.height); i++)
     for (j = resR.x; j < (resR.x + resR.width); j++)
      if (XPointInRegion (region, j, i))
       img->data[i*img->bytes_per_line + j] = transTable[(unsigned char) img->data[i*img->bytes_per_line + j]];     
  }
}

DQUANT void quantizeOnCmap(XImage * img, int imgCol, unsigned char *redCo, unsigned char *greenCo,
 unsigned char *blueCo, unsigned char *trTbl)
{
 int remainCol, i, j, min;

 remainCol = imgCol;
 for (i = 0; i < MAX_CELLS; i++)
   trTbl[i] = (unsigned char) i;
 i = j = ApData.ColorNo;
 while ((remainCol > 0) && (i < maxColors))
  {
   xinfo.allocCells[i].red = redCo[i - j] << 8;
   xinfo.allocCells[i].green = greenCo[i - j] << 8;
   xinfo.allocCells[i].blue = blueCo[i - j] << 8;
   xinfo.red[i] = redCo[i - j];
   xinfo.green[i] = greenCo[i - j];
   xinfo.blue[i] = blueCo[i - j];
   if (!xinfo.trueColor)
     XStoreColor (xinfo.dpy, xinfo.cmap, &xinfo.allocCells[i]);
   trTbl[xinfo.allocCells[i - j].pixel] = xinfo.allocCells[i].pixel;
   i++;
   remainCol--;
   ApData.ColorNo++;
  }
 i = imgCol - remainCol - 1;
 if (i < 0)
   i = 0;
 while (remainCol)
  {
   int dr, dg, db, ind;
   unsigned long dist;

   min = 0xfffff;
   ind = 0;
   for (j = 0; j < maxColors; j++)
    {
     dr = (unsigned long) (redCo[i] - xinfo.red[j]);
     dg = (unsigned long) (greenCo[i] - xinfo.green[j]);
     db = (unsigned long) (blueCo[i] - xinfo.blue[j]);
     dist = dr * dr + dg * dg + db * db;
     if (dist < min)
      { min = dist; ind = j; }
    }
   trTbl[xinfo.allocCells[i].pixel] = xinfo.allocCells[ind].pixel;
   i++;
   remainCol--;
  }
 remapData((unsigned char *) img->data, trTbl, img->bytes_per_line * img->height);

 for (i = 0; i < maxColors; i++)
  {
   redCo[i] = xinfo.red[i];
   greenCo[i] = xinfo.green[i];
   blueCo[i] = xinfo.blue[i];
  }
}

/* count the actual number of colors */
DQUANT int countHistColors(unsigned long *colHist)
{
 int i, numColors = 0;

 for (i = 0; i < MAX_CELLS; i++)
   if (colHist[i] > 0)
     numColors++;
 return(numColors);
}
