/* ltiff.c derived from David Koblas's xpaint */
#include "loadimg.h"
#include <stdio.h>
#include <stdlib.h>
#include "tiffio.h"

extern void registerLoader(tryLoadProc tLoad, loadBandProc lBand, endLoadProc eLoad);

static void nop(void)
{}

int TestTIFF(char *file)
{
 TIFF *tif;

 TIFFSetErrorHandler((TIFFErrorHandler) nop);
 if ((tif = TIFFOpen(file, "r")) == NULL)
     return 0;
 TIFFClose(tif);
 return 1;
}

static int forceRGB;
static unsigned short planarconfig;
static unsigned short *redmap, *greenmap, *bluemap;
static TIFF *tif;
static unsigned short bitsPS;
static unsigned short samplesPP;
static unsigned short photometric;
static unsigned short orientation;

static int width, height;
static int cmapSize;
static int readLines;

static int tryLoad(char *filename, int *loadType, int *xpixels,
 int *ypixels, int *xdpi, int *ydpi, unsigned char *cmap, char *info,
 char *reverse, int *numInterlace, int realLoad)
{
 int maxval, ncolors;
 int i;
 uint16 resUnit;
 float xres, yres;

 forceRGB = FALSE;

 TIFFSetErrorHandler((TIFFErrorHandler) nop);
 if ((tif = TIFFOpen(filename, "r")) == NULL)
  return(LOAD_FILETYPE_UNKNOWN);

 if (!TIFFGetField(tif, TIFFTAG_BITSPERSAMPLE, &bitsPS))
     bitsPS = 1;
 if (!TIFFGetField(tif, TIFFTAG_SAMPLESPERPIXEL, &samplesPP))
     samplesPP = 1;
 if (!TIFFGetField(tif, TIFFTAG_PHOTOMETRIC, &photometric))
  return(LOAD_FILETYPE_UNKNOWN);

 if (TIFFIsTiled(tif))
  return(LOAD_FILETYPE_UNKNOWN);
 if (samplesPP != 1 && samplesPP != 3 && samplesPP != 4)
  return(LOAD_FILETYPE_UNKNOWN);
 if (photometric == PHOTOMETRIC_YCBCR || photometric == PHOTOMETRIC_CIELAB ||
     photometric == PHOTOMETRIC_SEPARATED || photometric == PHOTOMETRIC_MASK)
  return(LOAD_FILETYPE_UNKNOWN);

 TIFFGetField(tif, TIFFTAG_IMAGEWIDTH, xpixels);
 TIFFGetField(tif, TIFFTAG_IMAGELENGTH, ypixels);
 TIFFGetFieldDefaulted(tif, TIFFTAG_PLANARCONFIG, &planarconfig);

 TIFFGetField(tif, TIFFTAG_RESOLUTIONUNIT, &resUnit);
 TIFFGetField(tif, TIFFTAG_XRESOLUTION, &xres);
 TIFFGetField(tif, TIFFTAG_YRESOLUTION, &yres);
 if (resUnit==RESUNIT_INCH)
  {
   *xdpi = (int) (xres*1000+0.5);
   *ydpi = (int) (yres*1000+0.5);
  }
 else
  if (resUnit==RESUNIT_CENTIMETER)
   {
    *xdpi = (int) (xres*1000/2.54+0.5);
    *ydpi = (int) (yres*1000/2.54+0.5);
   }
  else
   *xdpi = *ydpi = 0;

 if (planarconfig == PLANARCONFIG_SEPARATE && photometric != PHOTOMETRIC_RGB)
  return(LOAD_FILETYPE_UNKNOWN);

 ncolors = 1 << bitsPS;
 maxval = ncolors - 1;
 if (maxval == 1 && samplesPP == 1)
  {
   cmap[0]=cmap[1]=cmap[2]= (photometric == PHOTOMETRIC_MINISBLACK) ? 0 : 255;
   cmap[3]=cmap[4]=cmap[5]= (photometric == PHOTOMETRIC_MINISBLACK) ? 255 : 0;
   *loadType=LOAD_CMAP; /* note that BW is loaded as CMAP by this loader */
  }
 else
  switch (photometric)
   {
    case PHOTOMETRIC_MINISBLACK:
    case PHOTOMETRIC_MINISWHITE:
     for (i = 0; i < ncolors; i++)
      cmap[i*3]=cmap[i*3+1]=cmap[i*3+2]=
       (photometric==PHOTOMETRIC_MINISBLACK) ? (i*255)/maxval : ((maxval-i)*255)/maxval;
     *loadType=LOAD_CMAP; /* note that BW is loaded as CMAP by this loader */
     break;
    case PHOTOMETRIC_PALETTE:
     if (!TIFFGetField(tif, TIFFTAG_COLORMAP, &redmap, &greenmap, &bluemap))
      return(LOAD_FILETYPE_UNKNOWN);
     if (ncolors > 0xffff)
      {
       forceRGB = TRUE;
       *loadType=LOAD_TRUE_COL;
       break;
      }
     else
      for (i = 0; i < ncolors; i++)
       {
        cmap[i*3]=redmap[i]>>8;
        cmap[i*3+1]=greenmap[i]>>8;
        cmap[i*3+2]=bluemap[i]>>8;
       }
     *loadType=LOAD_CMAP;
     break;

     case PHOTOMETRIC_RGB:
      *loadType=LOAD_TRUE_COL;
      break;
    }
 if (!TIFFGetField(tif, TIFFTAG_ORIENTATION, &orientation))
     orientation = ORIENTATION_TOPLEFT;

 strcpy(info, "TIFF");

 width = *xpixels;
 height = *ypixels;
 cmapSize = ncolors;
 readLines = 0;

 switch (orientation)
  {
   default:
   case ORIENTATION_LEFTTOP:
   case ORIENTATION_TOPLEFT:
   case ORIENTATION_LEFTBOT:
   case ORIENTATION_BOTLEFT:
    *reverse = FALSE;
    break;
   case ORIENTATION_RIGHTTOP:
   case ORIENTATION_TOPRIGHT:
   case ORIENTATION_RIGHTBOT:
   case ORIENTATION_BOTRIGHT:
    *reverse = TRUE;
    break;
  }
 *numInterlace = 1; /* no interlace */
 return(LOAD_OK);
}

/* loader reads a band */
static int loadBand(unsigned char *data, int  lines, int  bpl)
{
 int row, rowStart, colStart, rowInc, colInc;
 int isWide;
 unsigned char *ucp;
 unsigned short *usp;
 unsigned char *sline = NULL, *sptr0=NULL, *sptr1=NULL, *sptr2=NULL;
 int x, y;
 int maxval = (1 << bitsPS) - 1;
 int scansize = 0;

 if (planarconfig == PLANARCONFIG_SEPARATE)
  {
   scansize = TIFFStripSize(tif);

   sline = (unsigned char *) malloc(3 * scansize);
   sptr0 = sline;
   sptr1 = sptr0 + scansize;
   sptr2 = sptr1 + scansize;
  }
 else
  if ((sline = (unsigned char *) malloc(TIFFScanlineSize(tif))) == NULL)
   return(LOAD_NO_MEMORY);

 isWide = cmapSize > 256;

 rowStart = 0;
 rowInc = 1;

 switch (orientation)
  {
   default:
   case ORIENTATION_LEFTTOP:
   case ORIENTATION_TOPLEFT:
   case ORIENTATION_LEFTBOT:
   case ORIENTATION_BOTLEFT:
    colStart = 0;
    colInc = 1;
    break;
   case ORIENTATION_RIGHTTOP:
   case ORIENTATION_TOPRIGHT:
   case ORIENTATION_RIGHTBOT:
   case ORIENTATION_BOTRIGHT:
    colStart = width;
    colInc = -1;
    break;
  }

 for (row = rowStart+readLines, y = 0; y < lines; y++, row += rowInc)
  {
   unsigned char *pp = sline;
   int bitsleft, sample0, sample1, sample2;

 usp = (unsigned short *) &data[bpl*y];
 ucp = (unsigned char *) &data[bpl*y];

#define NEXT3(p0, p1, p2) {			\
     if (bitsleft == 0) {			\
       p0++;					\
       p1++;					\
       p2++;					\
       bitsleft = 8;				\
     }					\
     bitsleft -= bitsPS;			\
     sample0 = (*p0 >> bitsleft) & maxval;	\
     sample1 = (*p1 >> bitsleft) & maxval;	\
     sample2 = (*p2 >> bitsleft) & maxval;	\
}

#define NEXT(pp) do {				\
     if (bitsleft == 0) {			\
       pp++;					\
       bitsleft = 8;				\
     }					\
     bitsleft -= bitsPS;			\
     sample0 = (*pp >> bitsleft) & maxval;	\
} while (0)

   if (planarconfig == PLANARCONFIG_SEPARATE)
    {
     if (TIFFReadEncodedStrip(tif, TIFFComputeStrip(tif, row, 0), sptr0, scansize) < 0)
      break;
     if (TIFFReadEncodedStrip(tif, TIFFComputeStrip(tif, row, 1), sptr1, scansize) < 0)
      break;
     if (TIFFReadEncodedStrip(tif, TIFFComputeStrip(tif, row, 2), sptr2, scansize) < 0)
      break;
    }
   else
    if (TIFFReadScanline(tif, sline, row, 0) < 0)
     break;

   //  If colInc == -1, then just reorder the samples.  

   bitsleft = 8;

   switch (photometric)
    {
     case PHOTOMETRIC_PALETTE:
      if (forceRGB)
       {
	for (x = 0; x < width; x++)
         {
	  NEXT(pp);
	  *ucp++ = redmap[sample0];
	  *ucp++ = greenmap[sample0];
	  *ucp++ = bluemap[sample0];
	 }
	break;
       }
       //	fall through 
   case PHOTOMETRIC_MINISWHITE:
   case PHOTOMETRIC_MINISBLACK:
    for (x = 0; x < width; x++)
     {
      NEXT(pp);
      if (isWide)
	  *usp++ = sample0;
      else
	  *ucp++ = sample0;
     }
    break;
   case PHOTOMETRIC_RGB:
       if (planarconfig == PLANARCONFIG_SEPARATE)
       {
	   unsigned char *p0 = sptr0, *p1 = sptr1, *p2 = sptr2;

	   for (x = 0; x < width; x++)
	   {
	       NEXT3(p0, p1, p2);
	       *ucp++ = sample0;
	       *ucp++ = sample1;
	       *ucp++ = sample2;
	   }
       }
       else
       {
	   for (x = 0; x < width; x++)
	   {
	       NEXT(pp);
	       *ucp++ = sample0;
	       NEXT(pp);
	       *ucp++ = sample0;
	       NEXT(pp);
	       *ucp++ = sample0;
	       if (samplesPP == 4)
		   NEXT(pp);
	   }
       }
       break;
   case PHOTOMETRIC_MASK:
       for (x = 0; x < width; x++)
       {
	   NEXT(pp);
	   *ucp++ = sample0 ? 1 : 0;
       }
       break;
   }
 }

 if (sline)
  free(sline);

 readLines += lines;
 return(LOAD_OK);
}

/* closes the input file and releases any temporary storage */
static int endLoad()
{
 TIFFClose(tif);
 return(LOAD_OK);
}

/* register TIFF image loader */
void registerLoadTIFF()
{
 registerLoader(*tryLoad, *loadBand, *endLoad);
}
